from typing import Any, Dict, Final, List, Tuple

from paddlex_hps_server import (
    BaseTritonPythonModel,
    app_common,
    logging,
    protocol,
    schemas,
    utils,
)
from paddlex_hps_server.storage import SupportsGetURL, create_storage

_DEFAULT_MAX_NUM_INPUT_IMGS: Final[int] = 10
_DEFAULT_MAX_OUTPUT_IMG_SIZE: Final[Tuple[int, int]] = (2000, 2000)


class TritonPythonModel(BaseTritonPythonModel):
    def initialize(self, args):
        super().initialize(args)
        self._context = {}
        self._context["file_storage"] = None
        self._context["return_img_urls"] = False
        self._context["max_num_input_imgs"] = _DEFAULT_MAX_NUM_INPUT_IMGS
        self._context["max_output_img_size"] = _DEFAULT_MAX_OUTPUT_IMG_SIZE
        if self._app_config.extra:
            if "file_storage" in self._app_config.extra:
                self._context["file_storage"] = create_storage(
                    self._app_config.extra["file_storage"]
                )
            if "return_img_urls" in self._app_config.extra:
                self._context["return_img_urls"] = self._app_config.extra[
                    "return_img_urls"
                ]
            if "max_num_input_imgs" in self._app_config.extra:
                self._context["max_num_input_imgs"] = self._app_config.extra[
                    "max_num_input_imgs"
                ]
            if "max_output_img_size" in self._app_config.extra:
                self._context["max_output_img_size"] = self._app_config.extra[
                    "max_output_img_size"
                ]
        if self._context["return_img_urls"]:
            file_storage = self._context["file_storage"]
            if not file_storage:
                raise ValueError(
                    "The file storage must be properly configured when URLs need to be returned."
                )
            if not isinstance(file_storage, SupportsGetURL):
                raise TypeError(f"{type(file_storage)} does not support getting URLs.")

    def get_input_model_type(self):
        return schemas.doc_preprocessor.InferRequest

    def get_result_model_type(self):
        return schemas.doc_preprocessor.InferResult

    def run(self, input, log_id):
        if input.fileType is None:
            if utils.is_url(input.file):
                try:
                    file_type = utils.infer_file_type(input.file)
                except Exception:
                    logging.exception("Failed to infer the file type")
                    return protocol.create_output_without_result(
                        422,
                        "The file type cannot be inferred from the URL. Please specify the file type explicitly.",
                        log_id=log_id,
                    )
            else:
                return protocol.create_output_without_result(
                    422,
                    "Unknown file type",
                    log_id=log_id,
                )
        else:
            file_type = "PDF" if input.fileType == 0 else "IMAGE"

        file_bytes = utils.get_raw_bytes(input.file)
        images, data_info = utils.file_to_images(
            file_bytes,
            file_type,
            max_num_imgs=self._context["max_num_input_imgs"],
        )

        result = list(
            self._pipeline(
                images,
                use_doc_orientation_classify=input.useDocOrientationClassify,
                use_doc_unwarping=input.useDocUnwarping,
            )
        )

        doc_pp_results: List[Dict[str, Any]] = []
        for i, (img, item) in enumerate(zip(images, result)):
            pruned_res = app_common.prune_result(item.json["res"])
            output_img = app_common.postprocess_image(
                item["output_img"],
                log_id,
                "output_img.png",
                file_storage=self._context["file_storage"],
                return_url=self._context["return_img_urls"],
                max_img_size=self._context["max_output_img_size"],
            )
            if self._app_config.visualize:
                vis_imgs = {
                    "input_img": img,
                    "doc_preprocessing_img": item.img["preprocessed_img"],
                }
                vis_imgs = app_common.postprocess_images(
                    vis_imgs,
                    log_id,
                    filename_template=f"{{key}}_{i}.jpg",
                    file_storage=self._context["file_storage"],
                    return_urls=self._context["return_img_urls"],
                    max_img_size=self._context["max_output_img_size"],
                )
            else:
                vis_imgs = {}
            doc_pp_results.append(
                dict(
                    outputImage=output_img,
                    prunedResult=pruned_res,
                    docPreprocessingImage=vis_imgs.get("doc_preprocessing_img"),
                    inputImage=vis_imgs.get("input_img"),
                )
            )

        return schemas.doc_preprocessor.InferResult(
            docPreprocessingResults=doc_pp_results,
            dataInfo=data_info,
        )
